from abc import ABC
from .type_ import Type


class BuiltInType(Type, ABC):

    def __init__(self, type_):
        self.type_ = type_

    def __call__(self, arg):
        return self.type_(arg)

    def recursive_children(self):
        return []

    @property
    def self_description(self):
        return f"{self.type_.__name__}"


class Str(BuiltInType):
    @property
    def simple_description(self):
        return "a string"

    @property
    def regex(self):
        return r'"([^"]*)"'


class Bool(BuiltInType):
    @property
    def simple_description(self):
        return "a boolean"

    @property
    def regex(self):
        return r"(true|false)"


class Float(BuiltInType):
    @property
    def simple_description(self):
        return "a float"

    @property
    def regex(self):
        return r"(\d+|\d*\.\d+(?!\d))"


class Int(BuiltInType):
    @property
    def simple_description(self):
        return "an integer"

    @property
    def regex(self):
        return r"(\d+)"



from .type_ import Type


class Class(Type):
    def __init__(self, cls):
        self.cls = cls
        self.kwargs = {
            name: Type(type_)
            for name, type_ in self.cls.__init__.__annotations__.items()
            if name != "return"
        }

    def recursive_children(self):
        children = [self]
        for type_ in self.kwargs.values():
            if isinstance(type_, Class):
                children.extend(type_.recursive_children())
        return children

    @property
    def path(self):
        return f"{self.cls.__module__}.{self.cls.__name__}"

    def kwarg_regex(self, kwarg):
        type_regex = self.kwargs[kwarg].regex

        return rf'"{kwarg}":{type_regex}'

    @property
    def kwargs_regex(self):
        string = ",".join(map(self.kwarg_regex, self.kwargs))
        return r"\{" + string + "\}"

    @property
    def regex(self):
        return r'\{"type":"' + self.path.replace(".", "\.") + r'","kwargs":' + self.kwargs_regex + r'\}'

    def __call__(self, response):
        kwargs = response["kwargs"]
        return self.cls(**{
            key: self.kwargs[key](value)
            for key, value in kwargs.items()
        })

    @property
    def self_description(self):
        return (
                f"A {self.cls.__name__} is a dictionary with a key 'type' and value '{self.path}' "
                + f"and a key 'kwargs'\n"
                + "The kwargs are:\n"
                + "\n".join(
            f"  - {kwarg}: {type_.simple_description}" for kwarg, type_ in self.kwargs.items()
        )
        )

    @property
    def simple_description(self):
        return f"{self.cls.__name__}"

    def __repr__(self):
        return f"Class({self.cls.__name__})"

    def __str__(self):
        return self.description



from typing import Union, Callable

from regex import regex
from rellm import complete_re
from transformers import PreTrainedTokenizer, PreTrainedModel

from orellm.type_ import Type


def complete_object(
        prompt: str, tokenizer: PreTrainedTokenizer,
        return_type: Union[Type, type],
        model: PreTrainedModel,
        max_new_tokens: int = 800,
        complete_function: Callable = complete_re,
        generate_prompt=False,
        **model_kwargs,
):
    """
    Create an instance of a class from a prompt

    Parameters
    ----------
    prompt
        A natural language prompt to give to the model
    tokenizer
    return_type
        The type that should be returned
    model
        A Large Language Model
    max_new_tokens
        The maximum number of tokens to generate
    complete_function
        A function that produces LLM output constrained by a regex pattern (e.g. complete_re from ReLLM)
    model_kwargs
        Additional arguments to pass to the model
    generate_prompt
        Whether to generate a prompt from the return type

    Returns
    -------
    An instance of the class specified by return_type
    """
    if not isinstance(return_type, Type):
        return_type = Type(return_type)

    prefix = return_type.description if generate_prompt else ""

    pattern = regex.compile(return_type.regex)
    response = complete_function(
        tokenizer=tokenizer,
        model=model,
        prompt=f"{prefix}\n\n{prompt}",
        pattern=pattern,
        max_new_tokens=max_new_tokens,
        do_sample=True,
        **model_kwargs
    )
    return return_type.from_json(response)



from .built_in_type import BuiltInType
from .type_ import Type
import typing


class Collection(BuiltInType):
    type: typing.Collection

    def __init__(self, child_type):
        if not isinstance(child_type, Type):
            child_type = Type(child_type)
        self.child_type = child_type
        super().__init__(self.type)

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    @property
    def regex(self):
        return f"\[({self.child_type.regex}(,{self.child_type.regex})*)?\]"

    def recursive_children(self):
        return [self, *self.child_type.recursive_children()]

    @property
    def simple_description(self):
        return f"a {self.type_.__name__} of {self.child_type.simple_description}"

    @property
    def self_description(self):
        return f"a {self.type_.__name__} of {self.child_type.simple_description}"

    def __call__(self, response):
        return super().__call__(map(self.child_type, response))


class List(Collection):
    type = list



class Simple:
    def __init__(self, argument: float):
        self.argument = argument

    def __eq__(self, other):
        if not isinstance(other, Simple):
            return False
        return self.argument == other.argument

    def __repr__(self):
        return f"Simple({self.argument})"


class Types:
    def __init__(
            self,
            int_argument: int,
            string_argument: str,
            boolean_argument: bool,
            float_argument: float,
    ):
        self.int_argument = int_argument
        self.string_argument = string_argument
        self.boolean_argument = boolean_argument
        self.float_argument = float_argument

    def __eq__(self, other):
        if not isinstance(other, Types):
            return False
        return (
                self.int_argument == other.int_argument
                and self.string_argument == other.string_argument
                and self.boolean_argument == other.boolean_argument
                and self.float_argument == other.float_argument
        )

    def __repr__(self):
        return f"Types({self.int_argument}, {self.string_argument}, {self.boolean_argument}, {self.float_argument})"


class Nested:
    def __init__(self, simple: Simple):
        self.simple = simple

    def __eq__(self, other):
        if not isinstance(other, Nested):
            return False
        return self.simple == other.simple

    def __repr__(self):
        return f"Nested({self.simple})"



import json
from abc import ABC, abstractmethod
from inspect import isclass


class Type(ABC):
    """
    Any type that might be returned by a call to the LLM
    """
    @property
    @abstractmethod
    def regex(self) -> str:
        """
        A regular expression used by ReLLM to constrain the output such that it can
        be parsed as this type
        """

    @abstractmethod
    def __call__(self, response) -> object:
        """
        Convert the response into the appropriate type. The response conforms to the
        regex defined by this type
        """

    def from_json(self, response: str) -> object:
        """
        Convert a JSON string returned by the model into the appropriate type
        """
        return self.__call__(json.loads(response))

    def __new__(cls, type_):
        """
        Wrap a type in a Type object.

        Only supports some built-in types and classes with type annotations

        Parameters
        ----------
        type_
            The type to wrap
        """
        from .built_in_type import Int, Float, Bool, Str
        from .collection import List
        from .class_ import Class

        type_map = {
            int: Int,
            float: Float,
            bool: Bool,
            str: Str,
            list: List
        }

        try:
            return object.__new__(type_map[type_])
        except KeyError:
            if isclass(type_):
                return object.__new__(Class)

        raise TypeError(f"Type {type_} not supported")

    @abstractmethod
    def recursive_children(self):
        pass

    @property
    @abstractmethod
    def simple_description(self):
        pass

    @property
    def description(self):
        return "\n\n".join(child.self_description for child in self.recursive_children())

    @property
    @abstractmethod
    def self_description(self):
        pass
